<?php

declare(strict_types=1);

namespace Symplify\PHPStanRules\Nette\Rules;

use PhpParser\Node;
use PhpParser\Node\Expr\Assign;
use PhpParser\Node\Expr\PropertyFetch;
use PhpParser\Node\Stmt\Class_;
use PhpParser\NodeFinder;
use PHPStan\Analyser\Scope;
use PHPStan\Node\InClassNode;
use PHPStan\Rules\Rule;
use PHPStan\Rules\RuleError;
use PHPStan\Rules\RuleErrorBuilder;
use Symplify\PHPStanRules\Nette\NetteInjectAnalyzer;
use Symplify\RuleDocGenerator\Contract\DocumentedRuleInterface;
use Symplify\RuleDocGenerator\ValueObject\CodeSample\CodeSample;
use Symplify\RuleDocGenerator\ValueObject\RuleDefinition;

/**
 * @see \Symplify\PHPStanRules\Tests\Nette\Rules\ForbiddenNetteInjectOverrideRule\ForbiddenNetteInjectOverrideRuleTest
 * @implements Rule<InClassNode>
 */
final class ForbiddenNetteInjectOverrideRule implements Rule
{
    /**
     * @var string
     */
    public const ERROR_MESSAGE = 'Assign to already injected property is not allowed';
    /**
     * @var \Symplify\PHPStanRules\Nette\NetteInjectAnalyzer
     */
    private $netteInjectAnalyzer;
    /**
     * @var \PhpParser\NodeFinder
     */
    private $nodeFinder;
    public function __construct(NetteInjectAnalyzer $netteInjectAnalyzer, NodeFinder $nodeFinder)
    {
        $this->netteInjectAnalyzer = $netteInjectAnalyzer;
        $this->nodeFinder = $nodeFinder;
    }

    /**
     * @return class-string<Node>
     */
    public function getNodeType(): string
    {
        return InClassNode::class;
    }

    /**
     * @param InClassNode $node
     * @return RuleError[]
     */
    public function processNode(Node $node, Scope $scope): array
    {
        $classReflection = $node->getClassReflection();

        $classLike = $node->getOriginalNode();
        if (! $classLike instanceof Class_) {
            return [];
        }

        $parentClassReflections = $classReflection->getParents();
        if ($parentClassReflections === []) {
            return [];
        }

        /** @var Assign[] $assigns */
        $assigns = $this->nodeFinder->findInstanceOf($classLike->getMethods(), Assign::class);
        if ($assigns === []) {
            return [];
        }

        $errorMessages = [];

        foreach ($assigns as $assign) {
            if (! $assign->var instanceof PropertyFetch) {
                continue;
            }

            $propertyFetch = $assign->var;
            if (! $this->netteInjectAnalyzer->isParentInjectPropertyFetch(
                $propertyFetch,
                $parentClassReflections
            )) {
                continue;
            }

            $errorMessages[] = RuleErrorBuilder::message(self::ERROR_MESSAGE)
                ->line($propertyFetch->getLine())
                ->build();
        }

        return $errorMessages;
    }

    public function getRuleDefinition(): RuleDefinition
    {
        return new RuleDefinition(self::ERROR_MESSAGE, [
            new CodeSample(
                <<<'CODE_SAMPLE'
use Nette\DI\Attributes\Inject;

abstract class AbstractParent
{
    /**
     * @var SomeType
     */
    #[Inject]
    public $someType;
}

final class SomeChild extends AbstractParent
{
    public function __construct(AnotherType $anotherType)
    {
        $this->someType = $anotherType;
    }
}
CODE_SAMPLE
                ,
                <<<'CODE_SAMPLE'
use Nette\DI\Attributes\Inject;

abstract class AbstractParent
{
    /**
     * @var SomeType
     */
    #[Inject]
    public $someType;
}

final class SomeChild extends AbstractParent
{
}
CODE_SAMPLE
            ),
        ]);
    }
}
